"""General AutoPLEX automation jobs."""

from collections.abc import Iterable
from dataclasses import field
from pathlib import Path

import numpy as np
from atomate2.common.schemas.phonons import ForceConstants, PhononBSDOSDoc
from atomate2.vasp.jobs.base import BaseVaspMaker
from jobflow import Flow, Response, job
from pymatgen.core.structure import Structure
from pymatgen.phonon.bandstructure import PhononBandStructure
from pymatgen.phonon.dos import PhononDos

from autoplex.benchmark.phonons.flows import PhononBenchmarkMaker
from autoplex.data.phonons.flows import (
    DFTPhononMaker,
    IsoAtomMaker,
    IsoAtomStaticMaker,
    MLPhononMaker,
    RandomStructuresDataGenerator,
    TightDFTStaticMaker,
)
from autoplex.data.phonons.jobs import reduce_supercell_size


@job(
    data=[
        PhononBSDOSDoc,
        "dft_references",
        "metrics",
        "benchmark_structures",
        PhononDos,
        PhononBandStructure,
        ForceConstants,
        Structure,
    ]
)
def do_iterative_rattled_structures(
    workflow_maker_gen_0,
    workflow_maker_gen_1,
    structure_list: list[Structure],
    mp_ids: list[str],
    dft_references: list[PhononBSDOSDoc] | None = None,
    benchmark_structures: list[Structure] | None = None,
    benchmark_mp_ids: list[str] | None = None,
    pre_xyz_files: list[str] | None = None,
    pre_database_dir: str | None = None,
    rattle_seed: int | None = None,
    fit_kwargs_list: list | None = None,
    number_of_iteration=0,
    rms=0.2,
    max_iteration=5,
    rms_max=0.2,
    previous_output=None,
):
    """
    Job to run CompleteDFTvsMLBenchmarkWorkflow in an iterative manner.

    Parameters
    ----------
    workflow_maker_gen_0: CompleteDFTvsMLBenchmarkWorkflow.
        First Iteration will be performed with this flow.
    workflow_maker_gen_1: CompleteDFTvsMLBenchmarkWorkflow.
        All Iterations after the first one will be performed with this flow.
    structure_list:
            List of pymatgen structures.
    mp_ids:
        Materials Project IDs.
    dft_references: list[PhononBSDOSDoc] | None
        List of DFT reference files containing the PhononBSDOCDoc object.
        Reference files have to refer to a finite displacement of 0.01.
        For benchmarking, only 0.01 is supported
    benchmark_structures: list[Structure] | None
        The pymatgen structure for benchmarking.
    benchmark_mp_ids: list[str] | None
        Materials Project ID of the benchmarking structure.
    pre_xyz_files: list[str] or None
        Names of the pre-database train xyz file and test xyz file.
    pre_database_dir: str or None
        The pre-database directory.
    rattle_seed: int | None
        Random seed.
    fit_kwargs_list : list[dict].
        Dict including MLIP fit keyword args.
    number_of_iteration: int
        Number of iterations.
    rms: float
        current maximum rms value
    max_iteration: int.
        Maximum number of iterations to run.
    rms_max: float.
        Will stop once the best potential has a max rmse below this value.
    previous_output: dict | None.
        Dict including the output of the previous flow.
    """
    if rms is None or (number_of_iteration < max_iteration and rms > rms_max):
        jobs = []

        if number_of_iteration == 0:
            workflow_maker = workflow_maker_gen_0
            job1 = workflow_maker_gen_0.make(
                structure_list=structure_list,
                mp_ids=mp_ids,
                dft_references=dft_references,
                benchmark_structures=benchmark_structures,
                benchmark_mp_ids=benchmark_mp_ids,
                pre_xyz_files=pre_xyz_files,
                pre_database_dir=pre_database_dir,
                rattle_seed=rattle_seed,
                fit_kwargs_list=fit_kwargs_list,
            )
        else:
            workflow_maker = workflow_maker_gen_1
            job1 = workflow_maker_gen_1.make(
                structure_list=structure_list,
                mp_ids=mp_ids,
                dft_references=dft_references,
                benchmark_structures=benchmark_structures,
                benchmark_mp_ids=benchmark_mp_ids,
                pre_xyz_files=pre_xyz_files,
                pre_database_dir=pre_database_dir,
                rattle_seed=rattle_seed,
                fit_kwargs_list=fit_kwargs_list,
            )

        # rms needs to be computed somehow
        job1.append_name("_" + str(number_of_iteration))
        jobs.append(job1)
        # order is the same as in the scaling "scale_cells"
        if workflow_maker.volume_custom_scale_factors is not None:
            rattle_seed = rattle_seed + (
                len(workflow_maker.volume_custom_scale_factors)
                * len(workflow_maker.structure_list)
            )
        elif workflow_maker.n_structures is not None:
            rattle_seed = rattle_seed + (workflow_maker.n_structures) * len(
                workflow_maker.structure_list
            )

        job2 = do_iterative_rattled_structures(
            workflow_maker_gen_0=workflow_maker_gen_0,
            workflow_maker_gen_1=workflow_maker_gen_1,
            structure_list=structure_list,
            mp_ids=mp_ids,
            dft_references=job1.output["dft_references"],
            benchmark_structures=job1.output["benchmark_structures"],
            benchmark_mp_ids=job1.output["benchmark_mp_ids"],
            pre_xyz_files=job1.output["pre_xyz_files"],
            pre_database_dir=job1.output["pre_database_dir"],
            rattle_seed=rattle_seed,
            fit_kwargs_list=fit_kwargs_list,
            number_of_iteration=number_of_iteration + 1,
            rms=job1.output["rms"],
            max_iteration=max_iteration,
            rms_max=rms_max,
            previous_output=job1.output,
        )
        jobs.append(job2)
        # recreate the output to make sure all is correctly put into data:
        output_dict = {
            "dft_references": job2.output["dft_references"],
            "benchmark_structures": job2.output["benchmark_structures"],
            "benchmark_mp_ids": job2.output["benchmark_mp_ids"],
            "pre_xyz_files": job2.output["pre_xyz_files"],
            "pre_database_dir": job2.output["pre_database_dir"],
            "rms": job2.output["rms"],
            "metrics": job2.output["metrics"],
            "fit_kwargs_list": job2.output["fit_kwargs_list"],
        }

        return Response(replace=Flow(jobs), output=output_dict)

    return {
        "dft_references": previous_output["dft_references"],
        "benchmark_structures": previous_output["benchmark_structures"],
        "benchmark_mp_ids": previous_output["benchmark_mp_ids"],
        "pre_xyz_files": previous_output["pre_xyz_files"],
        "pre_database_dir": previous_output["pre_database_dir"],
        "rms": previous_output["rms"],
        "metrics": previous_output["metrics"],
        "fit_kwargs_list": previous_output["fit_kwargs_list"],
    }


@job(data=[PhononBSDOSDoc])
def complete_benchmark(  # this function was put here to prevent circular import
    ml_path: list,
    ml_model: str,
    ibenchmark_structure: int,
    benchmark_structure: Structure,
    mp_ids,
    benchmark_mp_ids,  # list[str] mypy: Value of type "list[Any] | None" is not indexable
    add_dft_phonon_struct: bool,
    fit_input,
    symprec,
    phonon_bulk_relax_maker: BaseVaspMaker,
    phonon_static_energy_maker: BaseVaspMaker,
    phonon_displacement_maker: BaseVaspMaker,
    atomwise_regularization_parameter: float,
    dft_references=None,
    soap_dict=None,
    displacement: float = 0.01,
    supercell_settings: dict | None = None,
    relax_maker_kwargs: dict | None = None,
    static_maker_kwargs: dict | None = None,
    **ml_phonon_maker_kwargs,
):
    """
    Construct a complete flow for benchmarking the MLIP fit quality using a DFT based phonon structure.

    The complete benchmark flow starts by calculating the MLIP based phonon structure for each structure that has to be
    benchmarked.
    Then, depending on if the user provided a DFT reference dataset or the DFT reference structure is
    already given from a previous loop, the existing or to-be-calculated DFT reference is used to generate the phonon
    bandstructure comparison plots, the q-point wise RMSE plots and to calculate the overall RMSE.
    This process is repeated with the default ML potential as well as the potentials from the different active user
    settings like sigma regularization, separated datasets or looping through several sets of hyperparameters.

    Parameters
    ----------
    ml_path: str
        Path to MLIP file.
        Default is path to gap_file.xml
    ml_model: str
        ML model to be used.
        Default is GAP.
    ibenchmark_structure: int
        The ith benchmark structure.
    benchmark_structure: Structure
        The pymatgen structure for benchmarking.
    benchmark_mp_ids: list[str]
        Materials Project ID of the benchmarking structure.
    mp_ids:
        Materials Project IDs.
    add_dft_phonon_struct: bool.
        If True, will add displaced supercells via phonopy for DFT calculation.
    fit_input : dict.
        CompletePhononDFTMLDataGenerationFlow output.
    symprec: float
        Symmetry precision to use in the
        reduction of symmetry to find the primitive/conventional cell
        (use_primitive_standard_structure, use_conventional_standard_structure)
        and to handle all symmetry-related tasks in phonopy.
    phonon_bulk_relax_maker: BaseVaspMaker
        Maker used for the bulk relax unit cell calculation.
    phonon_static_energy_maker: BaseVaspMaker
        Maker used for the static energy unit cell calculation.
    phonon_displacement_maker: BaseVaspMaker
        Maker used to compute the forces for a supercell.
    dft_references:
        List of DFT reference files containing the PhononBSDOCDoc object.
        Default None.
    supercell_settings: dict
        Settings for supercell generation
    relax_maker_kwargs: dict
        Keyword arguments that can be passed to the RelaxMaker.
    static_maker_kwargs: dict
        Keyword arguments that can be passed to the StaticMaker.
    ml_phonon_maker_kwargs: dict
        Keyword arguments that can be passed to the MLPhononMaker.
    displacement: float
        Displacement used in the finite displacement method.
    atomwise_regularization_parameter: float
        Regularization value for the atom-wise force components.
    soap_dict: dict
        Dictionary containing SOAP parameters.

    """
    jobs = []
    collect_output = []

    for path in ml_path:
        suffix = Path(path).name
        if suffix == "without_regularization":
            suffix = "without_reg"
        if suffix not in ["phonon", "rattled"]:
            suffix = ""

        if phonon_displacement_maker is None:
            phonon_displacement_maker = TightDFTStaticMaker(name="dft phonon static")

        if ml_model == "GAP":
            ml_potential = (
                Path(path) / "gap_file.xml"
            )  # TODO account for user-specific gap file name?
        elif ml_model == "J-ACE":
            raise UserWarning("No atomate2 ACE.jl PhononMaker implemented.")
        elif ml_model in ["M3GNET"]:
            ml_potential = Path(
                path
            )  # M3GNet requires path and fit already returns the path
            # also need to find a different solution for separated fit then (name to path could be modified)
        elif ml_model in ["NEP"]:
            ml_potential = Path(path) / "nep.txt"
        elif ml_model in ["NEQUIP"]:
            ml_potential = Path(path) / "deployed_nequip_model.pth"
        else:  # MACE
            # treat finetuned potentials
            # TODO: fix this naming issue (depends on input)
            ml_potential_fine = Path(path) / "MACE_final.model"
            ml_potential = (
                ml_potential_fine
                if ml_potential_fine.exists()
                else Path(path) / "MACE_model.model"
            )
        if Path(ml_potential).exists():
            add_data_ml_phonon = MLPhononMaker(
                relax_maker_kwargs=relax_maker_kwargs,
                static_maker_kwargs=static_maker_kwargs,
                displacement=displacement,
            ).make_from_ml_model(
                structure=benchmark_structure,
                ml_model=ml_model,
                potential_file=ml_potential,
                supercell_settings=supercell_settings,
                **ml_phonon_maker_kwargs,
            )
            jobs.append(add_data_ml_phonon)

            # DFT benchmark reference preparations
            if dft_references is None and benchmark_mp_ids is not None:
                # runs only the first time, then dft_references is added
                if (
                    benchmark_mp_ids[ibenchmark_structure] in mp_ids
                ) and add_dft_phonon_struct:

                    dft_references = fit_input[benchmark_mp_ids[ibenchmark_structure]][
                        "phonon_data"
                    ][f"{int(displacement * 100):03d}"]
                else:
                    dft_phonons = dft_phonopy_gen_data(
                        structure=benchmark_structure,
                        mp_id=benchmark_mp_ids[ibenchmark_structure],
                        displacements=[displacement],
                        symprec=symprec,
                        phonon_bulk_relax_maker=phonon_bulk_relax_maker,
                        phonon_static_energy_maker=phonon_static_energy_maker,
                        phonon_displacement_maker=phonon_displacement_maker,
                        supercell_settings=supercell_settings,
                    )
                    jobs.append(dft_phonons)
                    dft_references = dft_phonons.output["phonon_data"][
                        f"{int(displacement * 100):03d}"
                    ]

                add_data_bm = PhononBenchmarkMaker(name="Benchmark").make(
                    ml_model=ml_model,
                    structure=benchmark_structure,
                    benchmark_mp_id=benchmark_mp_ids[ibenchmark_structure],
                    ml_phonon_task_doc=add_data_ml_phonon.output,
                    dft_phonon_task_doc=dft_references,
                    displacement=displacement,
                    atomwise_regularization_parameter=atomwise_regularization_parameter,
                    soap_dict=soap_dict,
                    suffix=suffix,
                )
            elif (
                dft_references is not None
                and not isinstance(dft_references, list)
                and benchmark_mp_ids is not None
            ):
                add_data_bm = PhononBenchmarkMaker(name="Benchmark").make(
                    # this is important for reusing the same internally calculated DFT reference
                    # for looping through several settings
                    ml_model=ml_model,
                    structure=benchmark_structure,
                    benchmark_mp_id=benchmark_mp_ids[ibenchmark_structure],
                    ml_phonon_task_doc=add_data_ml_phonon.output,
                    dft_phonon_task_doc=dft_references,
                    displacement=displacement,
                    atomwise_regularization_parameter=atomwise_regularization_parameter,
                    soap_dict=soap_dict,
                    suffix=suffix,
                )
            else:
                add_data_bm = PhononBenchmarkMaker(name="Benchmark").make(
                    # this is important for using a provided DFT reference
                    ml_model=ml_model,
                    structure=benchmark_structure,
                    benchmark_mp_id=benchmark_mp_ids[ibenchmark_structure],
                    ml_phonon_task_doc=add_data_ml_phonon.output,
                    dft_phonon_task_doc=dft_references[ibenchmark_structure],
                    displacement=displacement,
                    atomwise_regularization_parameter=atomwise_regularization_parameter,
                    soap_dict=soap_dict,
                    suffix=suffix,
                )
            jobs.append(add_data_bm)
            collect_output.append(add_data_bm.output)

    if isinstance(dft_references, list):
        return Response(
            replace=Flow(jobs),
            output={
                "bm_output": collect_output,
                "dft_references": dft_references[ibenchmark_structure],
            },
        )
    return Response(
        replace=Flow(jobs),
        output={"bm_output": collect_output, "dft_references": dft_references},
    )


@job
def generate_supercells(
    structures: list[Structure],
    supercell_settings: dict,
) -> list[Iterable]:
    """
    Run phonon displacements.

    Note, this job will replace itself with N displacement calculations,
    or a single socket calculation for all displacements.

    Parameters
    ----------
    structures : list[Structure]
        List of supercells.
    supercell_settings: dict
        Settings for supercells.

    """
    return [
        reduce_supercell_size(structure, **supercell_settings)
        for structure in structures
    ]


@job
def run_supercells(
    structures: list[Structure],
    supercell_matrices: list[int],
    mp_ids: list[str],
    dft_maker: BaseVaspMaker = None,
) -> Response:
    """
    Run supercell calculations.

    Note, this job will replace itself with supercell calculations.

    Parameters
    ----------
    structures: list[Structure]
        List of supercells.
    supercell_matrices: list[int]
        List of supercell matrices.
    mp_ids: list[str]
        List of Materials Project IDs.
    dft_maker : .BaseVaspMaker or .ForceFieldStaticMaker or .BaseAimsMaker
        Maker to use to generate dispacement calculations
    """
    if dft_maker is None:
        dft_maker = field(default_factory=TightDFTStaticMaker)
    dft_jobs = []
    outputs: dict[str, list] = {
        "uuids": [],
        "dirs": [],
        "mp-id": [],
    }

    for structure, supercell_matrix, mp_id in zip(
        structures, supercell_matrices, mp_ids
    ):
        structure.make_supercell(np.array(supercell_matrix).transpose())
        dft_job = dft_maker.make(structure=structure)
        dft_jobs.append(dft_job)
        outputs["uuids"].append(dft_job.output.uuid)
        outputs["dirs"].append(dft_job.output.dir_name)
        outputs["mp-id"].append(mp_id)

    displacement_flow = Flow(dft_jobs, outputs)
    return Response(replace=displacement_flow)


@job(data=["phonon_data"])
def dft_phonopy_gen_data(
    structure: Structure,
    mp_id: str,
    displacements,
    symprec,
    phonon_bulk_relax_maker,
    phonon_static_energy_maker,
    phonon_displacement_maker,
    supercell_settings,
):
    """
    Job to generate DFT reference database using phonopy to be used for fitting ML potentials.

    Parameters
    ----------
    structure: Structure
        The pymatgen Structure object.
    mp_id: str
        Materials Project ID.
    phonon_displacement_maker : .BaseVaspMaker or None
        Maker used to compute the forces for a supercell.
    phonon_bulk_relax_maker: BaseVaspMaker
        Maker used for the bulk relax unit cell calculation.
    phonon_static_energy_maker: BaseVaspMaker
        Maker used for the static energy unit cell calculation.
    displacements: list[float]
        List of phonon displacement.
    symprec : float
        Symmetry precision to use in the
        reduction of symmetry to find the primitive/conventional cell
        (use_primitive_standard_structure, use_conventional_standard_structure)
        and to handle all symmetry-related tasks in phonopy.
    supercell_settings:
        Settings for supercell generation.
    """
    jobs = []
    dft_phonons_output = {}
    dft_phonons_dir_output = []
    if supercell_settings is None:
        supercell_settings = field(
            default_factory=lambda: {"min_length": 15, "max_length": 20}
        )
    supercell_matrix = supercell_settings.get(mp_id, {}).get("supercell_matrix")
    if not supercell_matrix:
        filtered_settings = {  # mismatching mp_ids would lead to a key error
            key: value
            for key, value in supercell_settings.items()
            if key
            in [
                "min_length",
                "max_length",
                "fallback_min_length",
                "max_atoms",
                "min_atoms",
                "step_size",
            ]
        }
        supercell_matrix = reduce_supercell_size(structure, **filtered_settings)

    if phonon_displacement_maker is None:
        phonon_displacement_maker = TightDFTStaticMaker(name="dft phonon static")

    # always set autoplex default as job name
    phonon_displacement_maker.name = "dft phonon static"
    if phonon_static_energy_maker is not None:
        phonon_static_energy_maker.name = "dft static"
    if phonon_bulk_relax_maker is not None:
        try:
            phonon_bulk_relax_maker.relax_maker1.name = "dft tight relax"
            phonon_bulk_relax_maker.relax_maker2.name = "dft tight relax"
        except AttributeError:
            phonon_bulk_relax_maker.name = "dft tight relax"

    for displacement in displacements:
        dft_phonons = DFTPhononMaker(
            symprec=symprec,
            static_energy_maker=phonon_static_energy_maker,
            bulk_relax_maker=phonon_bulk_relax_maker,
            phonon_displacement_maker=phonon_displacement_maker,
            born_maker=None,
            displacement=displacement,
        ).make(structure=structure, supercell_matrix=supercell_matrix)
        jobs.append(dft_phonons)
        dft_phonons_output[
            f"{displacement}".replace(".", "")  # key must not contain '.'
        ] = dft_phonons.output
        dft_phonons_dir_output.append(dft_phonons.output.jobdirs.displacements_job_dirs)

    flow = Flow(
        jobs,
        {"phonon_dir": dft_phonons_dir_output, "phonon_data": dft_phonons_output},
        name="dft_phononpy_gen_data",
    )
    return Response(replace=flow)


@job
def dft_random_gen_data(
    structure: Structure,
    mp_id,
    rattled_bulk_relax_maker,
    displacement_maker,
    uc: bool = False,
    volume_custom_scale_factors: list[float] | None = None,
    volume_scale_factor_range: list[float] | None = None,
    rattle_std: float = 0.01,
    distort_type: int = 0,
    n_structures: int = 10,
    min_distance: float = 1.5,
    angle_percentage_scale: float = 10,
    angle_max_attempts: int = 1000,
    rattle_type: int = 0,
    rattle_seed: int = 42,
    rattle_mc_n_iter: int = 10,
    w_angle: list[float] | None = None,
    supercell_settings: dict | None = None,
):
    """
    Job to generate random structured DFT reference database to be used for fitting ML potentials.

    Parameters
    ----------
    structure: Structure
        The pymatgen Structure object
    displacement_maker : .BaseVaspMaker or None
        Maker used for a static calculation for a supercell.
    rattled_bulk_relax_maker: BaseVaspMaker
        Maker used for the bulk relax unit cell calculation.
    mp_id:
        Materials Project ID.
    uc: bool.
        If True, will generate randomly distorted structures (unitcells)
        and add static computation jobs to the flow.
    distort_type : int.
        0- volume distortion, 1- angle distortion, 2- volume and angle distortion. Default=0.
    distort_type : int.
        0- volume distortion, 1- angle distortion, 2- volume and angle distortion. Default=0.
    n_structures : int.
        Total number of distorted structures to be generated.
        Must be provided if distorting volume without specifying a range, or if distorting angles.
        Default=10.
    volume_scale_factor_range : list[float]
        [min, max] of volume scale factors.
        e.g. [0.90, 1.10] will distort volume +-10%.
    volume_custom_scale_factors : list[float]
        Specify explicit scale factors (if range is not specified).
        If None, will default to [0.90, 0.95, 0.98, 0.99, 1.01, 1.02, 1.05, 1.10].
    min_distance: float
        Minimum separation allowed between any two atoms.
        Default= 1.5A.
    angle_percentage_scale: float
        Angle scaling factor.
        Default= 10 will randomly distort angles by +-10% of original value.
    angle_max_attempts: int.
        Maximum number of attempts to distort structure before aborting.
        Default=1000.
    w_angle: list[float]
        List of angle indices to be changed i.e. 0=alpha, 1=beta, 2=gamma.
        Default= [0, 1, 2].
    rattle_type: int.
        0- standard rattling, 1- Monte-Carlo rattling. Default=0.
    rattle_std: float.
        Rattle amplitude (standard deviation in normal distribution).
        Default=0.01.
        Note that for MC rattling, displacements generated will roughly be
        rattle_mc_n_iter**0.5 * rattle_std for small values of n_iter.
    rattle_seed: int.
        Seed for setting up NumPy random state from which random numbers are generated.
        Default=42.
    rattle_mc_n_iter: int.
        Number of Monte Carlo iterations.
        Larger number of iterations will generate larger displacements.
        Default=10.
    supercell_settings: dict
        Settings for supercells.
    """
    jobs = []

    if displacement_maker is None:
        displacement_maker = TightDFTStaticMaker(name="dft rattle static")

    # always set autoplex default as job name
    displacement_maker.name = "dft rattle static"
    if rattled_bulk_relax_maker is not None:
        rattled_bulk_relax_maker.name = "dft tight relax"

    # TODO: decide if we should remove the additional response here as well
    # looks like only the output is changing
    random_datagen = RandomStructuresDataGenerator(
        name="RandomDataGen",
        bulk_relax_maker=rattled_bulk_relax_maker,
        displacement_maker=displacement_maker,
        n_structures=n_structures,
        uc=uc,
        rattle_std=rattle_std,
        distort_type=distort_type,
        min_distance=min_distance,
        rattle_seed=rattle_seed,
        rattle_type=rattle_type,
        angle_max_attempts=angle_max_attempts,
        angle_percentage_scale=angle_percentage_scale,
        rattle_mc_n_iter=rattle_mc_n_iter,
        w_angle=w_angle,
        supercell_settings=supercell_settings,
    ).make(
        structure=structure,
        mp_id=mp_id,
        volume_custom_scale_factors=volume_custom_scale_factors,
        volume_scale_factor_range=volume_scale_factor_range,
    )
    jobs.append(random_datagen)

    flow = Flow(jobs, {"rattled_dir": random_datagen.output})
    return Response(replace=flow)


@job
def get_iso_atom(
    structure_list: list[Structure],
    isolated_atom_maker: IsoAtomStaticMaker,
):
    """
    Job to collect all atomic species of the structures and starting VASP calculation of isolated atoms.

    Parameters
    ----------
    structure_list: list[Structure]
        List of pymatgen Structure objects
    isolated_atom_maker: IsoAtomStaticMaker
        VASP maker for the isolated atom calculation.
    """
    jobs = []
    iso_atoms_dict = {}
    all_species = list(
        {specie for s in structure_list for specie in s.types_of_species}
    )

    isoatoms = IsoAtomMaker().make(
        all_species=all_species,
        isolated_atom_maker=isolated_atom_maker,
    )
    jobs.append(isoatoms)

    for i, species in enumerate(all_species):
        iso_atoms_dict.update({species.number: isoatoms.output["energies"][i]})

    flow = Flow(
        jobs,
        {
            "species": all_species,
            "energies": iso_atoms_dict,
            "dirs": isoatoms.output["dirs"],
        },
    )
    return Response(replace=flow)


@job(data=[PhononBSDOSDoc, "dft_references"])
def get_phonon_output(
    metrics: list,
    benchmark_structures: list[Structure] | None = None,
    benchmark_mp_ids: list[str] | None = None,
    dft_references: list[PhononBSDOSDoc] | None = None,
    pre_xyz_files: list[str] | None = None,
    pre_database_dir: str | None = None,
    fit_kwargs_list: list | None = None,
):
    """
    Job to collect and process all phonon-related output information for a potential restart of the flow.

    This function aggregates benchmark results, DFT reference data, and other input parameters
    to determine the best fit RMSE from phonon calculations across all benchmark fits.

    Parameters
    ----------
    metrics: list[dict]
        List of metric dictionaries from complete_benchmark jobs.
    dft_references: list[PhononBSDOSDoc] | None
        List of DFT reference files containing the PhononBSDOCDoc object.
        Reference files have to refer to a finite displacement of 0.01.
        For benchmarking, only 0.01 is supported
    benchmark_structures: list[Structure] | None
        The pymatgen structure for benchmarking.
    benchmark_mp_ids: list[str] | None
        Materials Project ID of the benchmarking structure.
    pre_xyz_files: list[str] or None
        Names of the pre-database train xyz file and test xyz file.
    pre_database_dir: str or None
        The pre-database directory.
    fit_kwargs_list : list[dict].
        Dict including MLIP fit keyword args.
    """
    # TODO: potentially evaluation of imaginary modes
    try:
        rms_max_values = []  # get the largest rms in each fit

        for i in range(len(metrics[0])):
            rms_max_value = max(
                sublist[i]["benchmark_phonon_rmse"] for sublist in metrics
            )
            rms_max_values.append(rms_max_value)
        rms = min(rms_max_values)
    except TypeError:
        # Set a large value as a fall back if None is discovered
        rms = 1000.0
    return {
        "metrics": metrics,
        "rms": rms,  # get the best fit RMSE
        "benchmark_structures": benchmark_structures,
        "benchmark_mp_ids": benchmark_mp_ids,
        "dft_references": dft_references,
        "pre_xyz_files": pre_xyz_files,
        "pre_database_dir": pre_database_dir,
        "fit_kwargs_list": fit_kwargs_list,
    }



"""Utility functions for training data jobs."""

import logging
import os
import random
import warnings
from collections.abc import Iterable
from functools import partial
from itertools import chain
from multiprocessing import Pool
from pathlib import Path

import ase.io
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from ase import Atoms
from ase.atoms import Atom
from ase.calculators.singlepoint import SinglePointCalculator
from ase.io import Trajectory as AseTrajectory
from ase.io import write
from ase.units import GPa
from hiphive.structure_generation import generate_mc_rattled_structures
from pymatgen.core import Structure
from pymatgen.io.ase import AseAtomsAdaptor
from quippy import descriptors
from scipy.sparse.linalg import LinearOperator, svds
from sklearn.model_selection import StratifiedShuffleSplit
from threadpoolctl import threadpool_limits

from autoplex.fitting.common.regularization import (
    calculate_hull_nd,
    get_convex_hull,
    get_e_distance_to_hull,
    get_e_distance_to_hull_nd,
    label_stoichiometry_volume,
)

logging.basicConfig(level=logging.INFO, format="[%(levelname)s] %(message)s")


def flatten(atoms_object: Atoms | Iterable, recursive: bool = False) -> list[Atoms]:
    """
    Flatten an iterable fully, but excluding Atoms objects.

    Parameters
    ----------
    atoms_object: Atoms or Iterable
        An Atoms object or an iterable containing Atoms objects.
    recursive: bool
        If set to True, the function will recursively flatten the iterable.

    Returns
    -------
    A flattened list containing only Atoms objects.

    """
    iteration_list = []

    if recursive:
        for element in atoms_object:
            if isinstance(element, Iterable) and not isinstance(
                element, (str, bytes, ase.atoms.Atoms, ase.Atoms)
            ):
                iteration_list.extend(flatten(element, recursive=True))
            else:
                iteration_list.append(element)
        return iteration_list

    return [item for sublist in atoms_object for item in sublist]


def rms_dict(x_ref: np.ndarray | list, x_pred: np.ndarray | list) -> dict:
    """Compute RMSE and standard deviation of predictions with reference data.

    Adapted and adjusted from libatoms GAP tutorial page
    https://libatoms.github.io/GAP/gap_fitting_tutorial.html#make-simple-plots-of-the-energies-and-forces-on-the-EMT-and-GAP-datas

    Parameters
    ----------
    x_ref: np.ndarray.
        List of reference data.
    x_pred: np.ndarray.
        List of prediction.
    Note that x_ref and x_pred should be of same shape.

    Returns
    -------
    dict
        A dict with RMSE and std deviation of predictions.
    """
    x_ref = np.array(x_ref)
    x_pred = np.array(x_pred)
    if np.shape(x_pred) != np.shape(x_ref):
        raise ValueError("WARNING: not matching shapes in rms")
    error_2 = (x_ref - x_pred) ** 2
    average = np.sqrt(np.average(error_2))
    std_ = np.sqrt(np.var(error_2))
    return {"rmse": average, "std": std_}


def to_ase_trajectory(
    traj_obj, filename: str = "atoms.traj", store_magmoms=False
) -> AseTrajectory:  # adapted from ase
    """
    Convert to an ASE .Trajectory.

    Parameters
    ----------
    traj_obj:
        Trajectory object.
    filename : str | None
        Name of the file to write the ASE trajectory to.
        If None, no file is written.
    store_magmoms:
        The bool to store magnetic moments.
    """
    for idx in range(len(traj_obj["atom_positions"])):
        atoms = Atoms(symbols=list(traj_obj["atomic_number"]))  # .atoms.copy()
        atoms.set_positions(traj_obj["atom_positions"][idx])
        atoms.set_cell(traj_obj["cell"][idx])
        atoms.set_pbc("T T T")

        kwargs = {
            "energy": traj_obj["energy"][idx],
            "forces": traj_obj["forces"][idx],
            "stress": traj_obj["stresses"][idx],
        }
        if store_magmoms:
            kwargs["magmom"] = traj_obj.magmoms[idx]

        atoms.calc = SinglePointCalculator(atoms=atoms, **kwargs)
        with AseTrajectory(filename, "a" if idx > 0 else "w", atoms=atoms) as f:
            f.write()

    return AseTrajectory(filename, "r")


def scale_cell(
    structure: Structure,
    volume_scale_factor_range: list[float] | None = None,
    n_structures: int = 10,
    volume_custom_scale_factors: list[float] | None = None,
) -> list[Structure]:
    """
    Take in a pymatgen Structure object and generates stretched or compressed structures.

    Parameters
    ----------
    structure : Structure.
        Pymatgen structures object.
    volume_scale_factor_range : list[float]
        [min, max] of volume scale factors.
    n_structures : int.
        If specified a range, the number of structures to be generated with
        volume distortions equally spaced between min and max.
    volume_custom_scale_factors : list[float]
        Specify explicit scale factors (if range is not specified).
        If None, will default to [0.90, 0.95, 0.98, 0.99, 1.01, 1.02, 1.05, 1.10].

    Returns
    -------
    Response.output.
        Stretched or compressed structures.
    """
    atoms = AseAtomsAdaptor.get_atoms(structure)
    distorted_cells = []

    if volume_custom_scale_factors is not None:
        scale_factors_defined = volume_custom_scale_factors
        warnings.warn("Using your custom lattice scale factors", stacklevel=2)
    if volume_custom_scale_factors is None:
        if volume_scale_factor_range is None:
            volume_scale_factor_range = [0.90, 1.1]

        scale_factors_defined = np.arange(
            volume_scale_factor_range[0],
            volume_scale_factor_range[1]
            + (volume_scale_factor_range[1] - volume_scale_factor_range[0])
            / (n_structures - 1),
            (volume_scale_factor_range[1] - volume_scale_factor_range[0])
            / (n_structures - 1),
        )

        if not np.isclose(scale_factors_defined, 1.0).any():
            scale_factors_defined = np.append(scale_factors_defined, 1)
            scale_factors_defined = np.sort(scale_factors_defined)

        warnings.warn(
            f"Generated lattice scale factors {scale_factors_defined} within your range",
            stacklevel=2,
        )

    for scale_factor in scale_factors_defined:
        # make copy of ground state
        cell = atoms.copy()
        # set lattice parameter scale factor
        lattice_scale_factor = scale_factor ** (1 / 3)
        # scale cell volume and atomic positions
        cell.set_cell(lattice_scale_factor * atoms.get_cell(), scale_atoms=True)
        # store scaled cell
        distorted_cells.append(AseAtomsAdaptor.get_structure(cell))
    return distorted_cells


def check_distances(structure: Structure, min_distance: float = 1.5) -> bool:
    """
    Take in a pymatgen Structure object and check minimum distances between atoms using minimum image convention.

    Useful after distorting cell angles and rattling to check atoms aren't too close.

    Parameters
    ----------
    structure : Structure.
        Pymatgen structures object.
    min_distance: float
        Minimum separation allowed between any two atoms. Default= 1.5A.

    Returns
    -------
    Response.output.
        "True" if atoms are sufficiently spaced out i.e. all pairwise interatomic distances > min_distance.
    """
    atoms = AseAtomsAdaptor.get_atoms(structure)

    for i in range(len(atoms)):
        indices = [j for j in range(len(atoms)) if j != i]
        distances = atoms.get_distances(i, indices, mic=True)
        for distance in distances:
            if distance < min_distance:
                warnings.warn("Atoms too close.", stacklevel=2)
                return False
    return True


def random_vary_angle(
    structure: Structure,
    min_distance: float = 1.5,
    angle_percentage_scale: float = 10,
    w_angle: list[float] | None = None,
    n_structures: int = 8,
    angle_max_attempts: int = 1000,
) -> list[Structure]:
    """
    Take in a pymatgen Structure object and generates angle-distorted structures.

    Parameters
    ----------
    structure : Structure.
        Pymatgen structures object.
    min_distance: float
        Minimum separation allowed between atoms. Default= 1.5A.
    angle_percentage_scale: float
        Angle scaling factor.
        Default= 10 will randomly distort angles by +-10% of original value.
    w_angle: list[float]
        List of angle indices to be changed i.e. 0=alpha, 1=beta, 2=gamma.
        Default= [0, 1, 2].
    n_structures: int.
        Number of angle-distorted structures to generate.
    angle_max_attempts: int.
        Maximum number of attempts to distort structure before aborting.
        Default=1000.

    Returns
    -------
    Response.output.
        Angle-distorted structures.
    """
    atoms = AseAtomsAdaptor.get_atoms(structure)
    distorted_angle_cells = []
    generated_structures = 0  # counter to keep track of generated structures

    if w_angle is None:
        w_angle = [0, 1, 2]

    while generated_structures < n_structures:
        attempts = 0
        # make copy of ground state
        atoms_copy = atoms.copy()

        # stretch lattice parameters by 3% before changing angles
        # helps atoms to not be too close
        distorted_cells = scale_cell(
            AseAtomsAdaptor.get_structure(atoms_copy),
            volume_custom_scale_factors=[1.03],
        )

        distorted_supercells: Atoms = AseAtomsAdaptor.get_atoms(distorted_cells[0])

        # getting stretched supercell out of array
        newcell = distorted_supercells.cell.cellpar()

        # current angles
        alpha = atoms_copy.cell.cellpar()[3]
        beta = atoms_copy.cell.cellpar()[4]
        gamma = atoms_copy.cell.cellpar()[5]

        # convert angle distortion scale
        angle_percentage_scale = angle_percentage_scale / 100
        min_scale = 1 - angle_percentage_scale
        max_scale = 1 + angle_percentage_scale

        while attempts < angle_max_attempts:
            attempts += 1
            # new random angles within +-10% (default) of current angle
            new_alpha = random.randint(int(alpha * min_scale), int(alpha * max_scale))
            new_beta = random.randint(int(beta * min_scale), int(beta * max_scale))
            new_gamma = random.randint(int(gamma * min_scale), int(gamma * max_scale))

            newvalues = [new_alpha, new_beta, new_gamma]

            for wang, newv in zip(w_angle, newvalues):
                newcell[wang + 3] = newv

            # converting newcell back into an Atoms object so future functions work
            # scaling atoms to new distorted cell
            atoms_copy.set_cell(newcell, scale_atoms=True)

            # if successful structure generated, i.e. atoms are not too close, then break loop
            if check_distances(AseAtomsAdaptor.get_structure(atoms_copy), min_distance):
                # store scaled cell
                distorted_angle_cells.append(AseAtomsAdaptor.get_structure(atoms_copy))
                generated_structures += 1
                break  # break the inner loop if successful
        else:
            raise RuntimeError(
                "Maximum attempts (1000) reached without distorting angles successfully."
            )

    return distorted_angle_cells


def std_rattle(
    structure: Structure,
    n_structures: int = 5,
    rattle_std: float = 0.01,
    rattle_seed: int = 42,
) -> list[Structure]:
    """
    Take in a pymatgen Structure object and generates rattled structures.

    Uses standard ASE rattle.

    Parameters
    ----------
    structure : Structure.
        The pymatgen structures object.
    n_structures: int.
        Number of rattled structures to generate.
    rattle_std: float.
        Rattle amplitude (standard deviation in normal distribution). Default=0.01.
    rattle_seed: int.
        Seed for setting up NumPy random state from which random numbers are generated. Default= 42.

    Returns
    -------
    Response.output.
        Rattled structures.
    """
    atoms = AseAtomsAdaptor.get_atoms(structure)
    rattled_xtals = []
    for i in range(n_structures):
        if i == 0:
            copy = atoms.copy()
            copy.rattle(stdev=rattle_std, seed=rattle_seed)
            rattled_xtals.append(AseAtomsAdaptor.get_structure(copy))
        if i > 0:
            rattle_seed = rattle_seed + 1
            copy = atoms.copy()
            copy.rattle(stdev=rattle_std, seed=rattle_seed)
            rattled_xtals.append(AseAtomsAdaptor.get_structure(copy))
    return rattled_xtals


def mc_rattle(
    structure: Structure,
    n_structures: int = 5,
    rattle_std: float = 0.003,
    min_distance: float = 1.5,
    rattle_seed: int = 42,
    rattle_mc_n_iter: int = 10,
) -> list[Structure]:
    """
    Take in a pymatgen Structure object and generates rattled structures.

    Randomly draws displacements with a MC trial step that penalises displacements leading to very small interatomic
    distances.
    Displacements generated will roughly be n_iter**0.5 * rattle_std for small values of n_iter.
    See https://hiphive.materialsmodeling.org/moduleref/structures.html?highlight=generate_mc_rattled_structures for
    more details.

    Parameters
    ----------
    structure : Structure.
        The pymatgen structures object.
    n_structures: int.
        Number of rattled structures to generate.
    rattle_std: float.
        Rattle amplitude (standard deviation in normal distribution). N.B. this value is not connected to the final
        average displacement for the structures. Default= 0.003.
    min_distance: float.
        Minimum separation of any two atoms in the rattled structures. Used for computing the probability for each
        rattle move. Default= 1.5A.
    rattle_seed: int.
        Seed for setting up NumPy random state from which random numbers are generated. Default= 42.
    rattle_mc_n_iter: int.
        Number of Monte Carlo iterations. Larger number of iterations will generate larger displacements. Default=10.

    Returns
    -------
    Response.output.
        Monte-Carlo rattled structures.
    """
    atoms = AseAtomsAdaptor.get_atoms(structure)
    mc_rattle = generate_mc_rattled_structures(
        atoms=atoms,
        n_structures=n_structures,
        rattle_std=rattle_std,
        d_min=min_distance,
        seed=rattle_seed,
        n_iter=rattle_mc_n_iter,
    )
    return [AseAtomsAdaptor.get_structure(xtal) for xtal in mc_rattle]


def filter_outlier_energy(
    in_file: str, out_file: str, criteria: float = 0.0005
) -> None:
    """
    Filter data outliers via energy criteria and write them into files.

    Parameters
    ----------
    in_file:
        Reference file (e.g. DFT).
    out_file:
        MLIP generated data file.
    criteria:
        Energy filter threshold.

    """
    # read files
    in_atoms = ase.io.read(in_file, ":")
    for atoms in in_atoms:
        kwargs = {
            "energy": atoms.info["REF_energy"],
        }
        atoms.calc = SinglePointCalculator(atoms=atoms, **kwargs)
    out_atoms = ase.io.read(out_file, ":")

    atoms_in = []
    atoms_out = []
    outliers = []

    for at_in, at_out in zip(in_atoms[:-1], out_atoms):
        en_in = at_in.get_potential_energy() / len(at_in.get_chemical_symbols())
        en_out = at_out.get_potential_energy() / len(at_out.get_chemical_symbols())
        en_error = rms_dict(en_in, en_out)
        if abs(en_error["rmse"]) < criteria:
            atoms_in.append(at_in)
            atoms_out.append(at_out)
        else:
            outliers.append(at_in)

    output_files = {
        "filtered_in_energy": atoms_in,
        "filtered_out_energy": atoms_out,
        "outliers_energy": outliers,
    }

    # Iterate over output files and write them
    for suffix, atom_list in output_files.items():
        path = Path(in_file).with_name(Path(in_file).name.replace("train", suffix))
        path.parent.mkdir(parents=True, exist_ok=True)  # Ensure directory exists
        write(path, atom_list, append=True)


def filter_outlier_forces(
    in_file: str, out_file: str, symbol="Si", criteria: float = 0.1
) -> None:
    """
    Filter data outliers via force criteria and write them into files.

    Parameters
    ----------
    in_file:
        Reference file (e.g. DFT).
    out_file:
        MLIP generated data file.
    symbol:
        Chemical symbol.
    criteria:
        Force filter threshold.

    """
    # read files
    in_atoms = ase.io.read(in_file, ":")
    for atoms in in_atoms[:-1]:
        kwargs = {
            "forces": atoms.arrays["REF_forces"],
        }
        atoms.calc = SinglePointCalculator(atoms=atoms, **kwargs)
    out_atoms = ase.io.read(out_file, ":")
    atoms_in = []
    atoms_out = []
    outliers = []
    # extract data for only one species
    for at_in, at_out in zip(in_atoms[:-1], out_atoms):
        # get the symbols
        sym_all = at_in.get_chemical_symbols()
        # add force for each atom
        force_error = []
        for j, sym in enumerate(sym_all):
            if sym in symbol:
                in_force = at_in.get_forces()[j]
                out_force = at_out.arrays["force"][j]
                rms = rms_dict(in_force, out_force)
                force_error.append(rms["rmse"])
        at_in.info["max RMSE"] = max(force_error) if force_error else 0
        at_in.info["avg RMSE"] = (
            sum(force_error) / len(force_error) if force_error else 0
        )
        at_in.info["RMSE"] = force_error
        if not any(np.any(value > criteria) for value in force_error):
            atoms_in.append(at_in)
            atoms_out.append(at_out)
        else:
            outliers.append(at_in)

    output_files = {
        "filtered_in_force": atoms_in,
        "filtered_out_force": atoms_out,
        "outliers_force": outliers,
    }

    # Iterate over output files and write them
    for suffix, atom_list in output_files.items():
        path = Path(in_file).with_name(Path(in_file).name.replace("train", suffix))
        path.parent.mkdir(parents=True, exist_ok=True)  # Ensure directory exists
        write(path, atom_list, append=True)


def energy_plot(
    in_file: str | Path,
    out_file: str | Path,
    ax: plt.Axes,
    title: str = "Plot of energy",
    label: str = "energy",
) -> float:
    """
    Plot the distribution of energy per atom on the output vs the input.

    Adapted and adjusted from libatoms GAP tutorial page
    https://libatoms.github.io/GAP/gap_fitting_tutorial.html#make-simple-plots-of-the-energies-and-forces-on-the-EMT-and-GAP-datas

    Parameters
    ----------
    in_file:
        Reference file (e.g. DFT).
    out_file:
        MLIP generated data file.
    ax:
        Panel position in plt.subplots.
    title:
        Title of the plot.
    label:
        Legend label

    """
    # read files
    in_atoms = ase.io.read(in_file, ":")
    out_atoms = ase.io.read(out_file, ":")
    atoms_in: list = []
    atoms_out: list = []
    for at_in, at_out in zip(in_atoms, out_atoms):
        kwargs = {
            "energy": at_in.info["REF_energy"],
        }
        at_in.calc = SinglePointCalculator(atoms=at_in, **kwargs)
        for at, atoms in [(at_in, atoms_in), (at_out, atoms_out)]:
            if at.info["config_type"] not in {"isolated_atom", "IsolatedAtom"}:
                atoms.append(at)
    # list energies
    ener_in = [
        at.get_potential_energy() / len(at.get_chemical_symbols()) for at in atoms_in
    ]
    ener_out = [
        at.get_potential_energy() / len(at.get_chemical_symbols()) for at in atoms_out
    ]
    # scatter plot of the data
    ax.scatter(ener_in, ener_out, label=label)
    # get the appropriate limits for the plot
    for_limits = np.array(ener_in + ener_out)
    elim = (for_limits.min() - 0.01, for_limits.max() + 0.01)
    ax.set_xlim(elim)
    ax.set_ylim(elim)
    # add line of slope 1 for reference
    ax.plot(elim, elim, c="k")
    # set labels
    ax.set_ylabel("energy by GAP / eV")
    ax.set_xlabel("energy by DFT / eV")
    # set title
    ax.set_title(title)
    # set legend
    ax.legend(loc="upper right")
    # add text about RMSE
    _rms = rms_dict(ener_in, ener_out)
    rmse_text = (
        "RMSE:\n"
        + str(np.round(_rms["rmse"], 7))
        + " +- "
        + str(np.round(_rms["std"], 9))
        + "eV/atom"
    )
    ax.text(
        0.9,
        0.1,
        rmse_text,
        transform=ax.transAxes,
        fontsize="large",
        horizontalalignment="right",
        verticalalignment="bottom",
    )

    return _rms["rmse"]


def force_plot(
    in_file: str | Path,
    out_file: str | Path,
    ax: plt.Axes,
    symbol: str = "Si",
    title: str = "Plot of force",
    label: str = "force for ",
) -> float:
    """
    Plot the distribution of force components per atom on the output vs the input.

    Only plots for the given atom type(s).

    Adapted and adjusted from libatoms GAP tutorial page
    https://libatoms.github.io/GAP/gap_fitting_tutorial.html#make-simple-plots-of-the-energies-and-forces-on-the-EMT-and-GAP-datas

    Parameters
    ----------
    in_file:
        Reference file (e.g. DFT).
    out_file:
        MLIP generated data file.
    ax:
        Panel position in plt.subplots.
    symbol:
        Chemical symbol.
    title:
        Title of the plot.
    label:
        Legend label

    """
    in_atoms = ase.io.read(in_file, ":")

    for atoms in in_atoms[:-1]:
        kwargs = {
            "forces": atoms.arrays["REF_forces"],
        }
        atoms.calc = SinglePointCalculator(atoms=atoms, **kwargs)
    out_atoms = ase.io.read(out_file, ":")
    # extract data for only one species
    in_force, out_force = [], []
    for at_in, at_out in zip(in_atoms[:-1], out_atoms):
        # get the symbols
        sym_all = at_in.get_chemical_symbols()
        # add force for each atom
        for j, sym in enumerate(sym_all):
            if sym in symbol:
                in_force.append(at_in.get_forces()[j])
                # out_force.append(at_out.get_forces()[j]) \
                out_force.append(
                    at_out.arrays["force"][j]
                )  # because QUIP and ASE use different names
    # convert to np arrays, much easier to work with
    # in_force = np.array(in_force)
    # out_force = np.array(out_force)
    # scatter plot of the data
    ax.scatter(in_force, out_force, label=label + symbol)
    # get the appropriate limits for the plot
    for_limits = np.array(in_force + out_force)
    flim = (for_limits.min() - 1, for_limits.max() + 1)
    ax.set_xlim(flim)
    ax.set_ylim(flim)
    # add line of
    ax.plot(flim, flim, c="k")
    # set labels
    ax.set_ylabel("force by GAP / (eV/Å)")
    ax.set_xlabel("force by DFT / (eV/Å)")
    # set title
    ax.set_title(title)
    # set legend
    ax.legend(loc="upper right")
    # add text about RMSE
    _rms = rms_dict(in_force, out_force)
    rmse_text = (
        "RMSE:\n"
        + str(np.round(_rms["rmse"], 3))
        + " +- "
        + str(np.round(_rms["std"], 5))
        + "eV/Å"
    )
    ax.text(
        0.9,
        0.1,
        rmse_text,
        transform=ax.transAxes,
        fontsize="large",
        horizontalalignment="right",
        verticalalignment="bottom",
    )

    return _rms["rmse"]


def plot_energy_forces(
    title: str,
    energy_limit: float,
    force_limit: float,
    species_list: list | None = None,
    train_name: str = "train.extxyz",
    test_name: str = "test.extxyz",
) -> None:
    """
    Plot energy and forces of the data.

    Adapted and adjusted from libatoms GAP tutorial page
    https://libatoms.github.io/GAP/gap_fitting_tutorial.html#make-simple-plots-of-the-energies-and-forces-on-the-EMT-and-GAP-datas

    Parameters
    ----------
    title: str
        Title of the plot.
    energy_limit: float
        Energy limit for data filtering.
    force_limit: list
        Force limit for data filtering.
    species_list: str
        List of species.
    train_name: str
        Name of the training data file.
    test_name: str
        Name of the test data file.
    """
    quip_train_file = train_name.replace("train", "quip_train")
    quip_test_file = test_name.replace("test", "quip_test")
    path = Path(train_name)
    if species_list is None:
        species_list = ["Si"]
    fig, ax_list = plt.subplots(nrows=3, ncols=2, gridspec_kw={"hspace": 0.3})
    fig.set_size_inches(10, 15)
    ax_list = ax_list.flat[:]
    rmse = ["Energy and forces and train and test data\n"]

    pretty_species_list = (
        str(species_list).replace("['", "").replace("']", "").replace("'", "")
    )

    energy_rmse_train = energy_plot(
        train_name, quip_train_file, ax_list[0], "Energy on training data"
    )
    rmse.append(f"Energy train: {energy_rmse_train}")
    for species in species_list:
        force_rmse_train = force_plot(
            train_name,
            quip_train_file,
            ax_list[1],
            species,
            f"Force on training data - {pretty_species_list}",
        )
        rmse.append(f"Force train {species}: {force_rmse_train}")
    energy_rmse_test = energy_plot(
        test_name, quip_test_file, ax_list[2], "Energy on test data"
    )
    rmse.append(f"Energy test: {energy_rmse_test}")

    filter_outlier_energy(train_name, quip_train_file, energy_limit)
    # filter_outlier_energy(test_name, quip_test_file, energy_limit)
    for species in species_list:
        force_rmse_test = force_plot(
            test_name,
            quip_test_file,
            ax_list[3],
            species,
            f"Force on test data - {pretty_species_list}",
        )
        rmse.append(f"Force test {species}: {force_rmse_test}")
        filter_outlier_forces(train_name, quip_train_file, species, force_limit)
        # filter_outlier_forces(test_name, quip_test_file, species, force_limit)

    energy_plot(
        path.with_name(path.name.replace("train", "filtered_in_energy")),
        path.with_name(path.name.replace("train", "filtered_out_energy")),
        ax_list[4],
        "Energy on filtered data",
        "energy-filtered data, energy",
    )
    for species in species_list:
        force_plot(
            path.with_name(path.name.replace("train", "filtered_in_force")),
            path.with_name(path.name.replace("train", "filtered_out_force")),
            ax_list[5],
            species,
            f"Force on filtered data - {pretty_species_list}",
            "energy-filtered data, force for ",
        )
    energy_plot(
        path.with_name(path.name.replace("train", "filtered_in_energy")),
        path.with_name(path.name.replace("train", "filtered_out_energy")),
        ax_list[4],
        "Energy on filtered data",
        "force-filtered data, energy",
    )
    for species in species_list:
        force_plot(
            path.with_name(path.name.replace("train", "filtered_in_force")),
            path.with_name(path.name.replace("train", "filtered_out_force")),
            ax_list[5],
            species,
            f"Force on filtered data - {pretty_species_list}",
            "force-filtered data, force for ",
        )

    fig.suptitle(title, fontsize=16)
    plt.savefig(train_name.replace("train", "energy_forces").replace(".extxyz", ".png"))

    with open("energy_train_rmse.txt", "a") as file:
        for entry in rmse:
            file.write(f"{entry}\n")


class ElementCollection:
    """
    A class to handle different species operations for a collection of atoms.

    The Species class provides methods to extract unique chemical elements (species),
    determine all possible pairs of these species, and retrieve their atomic numbers
    in a formatted string.
    """

    def __init__(self, atoms):
        self.atoms = atoms

    def get_species(self) -> list:
        """Extract a list of unique species (chemical elements) from the atoms."""
        sepcies_list = []

        for at in self.atoms:
            sym_all = at.get_chemical_symbols()
            syms = list(set(sym_all))
            for sym in syms:
                if sym in sepcies_list:
                    continue
                sepcies_list.append(sym)

        return sepcies_list

    def find_element_pairs(self, symb_list=None) -> list:
        """
        Generate a list of all possible unique pairs of species.

        It can operate on an optional list of symbols or default to
        using the species extracted from the atoms.

        Parameters
        ----------
        symb_list:
            List of chemical symbols.

        """
        species_list = self.get_species() if symb_list is None else symb_list

        pairs = []

        for i in range(len(species_list)):
            for j in range(i, len(species_list)):
                pair = (species_list[i], species_list[j])
                pairs.append(pair)

        return pairs

    def get_number_of_species(self) -> int:
        """Return the number of unique species present among the atoms."""
        return len(self.get_species())

    def get_species_Z(self) -> str:
        """Return a formatted string of atomic numbers of the unique species."""
        atom_numbers = []
        for atom_type in self.get_species():
            atom = Atoms(atom_type, [(0, 0, 0)])
            atom_numbers.append(int(atom.get_atomic_numbers()[0]))

        species_Z = "{"
        for i in range(len(atom_numbers) - 1):
            species_Z += str(atom_numbers[i]) + " "
        species_Z += str(atom_numbers[-1]) + "}"

        return species_Z


def parallel_calc_descriptor_vec(atom: Atoms, selected_descriptor: str) -> Atoms:
    """
    Calculate the SOAP descriptor vector for a given atom and hypers in parallel.

    Parameters
    ----------
    atom : ase.Atoms
        The atom for which to calculate the descriptor vector.
    selected_descriptor : str
        The quip descriptor string to use for the calculation.

    Returns
    -------
    ase.Atom
        The input atoms, with the descriptor vector added to its info dictionary.

    Notes
    -----
    The descriptor vector is added to the atom's info dictionary with the key 'descriptor_vec'.
    """
    desc_object = descriptors.Descriptor(selected_descriptor)
    atom.info["descriptor_vec"] = desc_object.calc(atom)["data"]

    return atom


def cur_select(
    atoms,
    selected_descriptor,
    kernel_exp,
    select_nums,
    stochastic=True,
    random_seed=None,
) -> list[Atoms] | None:
    """
    Perform CUR selection on a set of atoms to get representative SOAP descriptors.

    Parameters
    ----------
    atoms: list of ase.Atoms
        The atoms for which to perform CUR selection.
    selected_descriptor: str
        The quip descriptor string to use for the calculation.
    kernel_exp: float
        The kernel exponent to use in the calculation.
    select_nums: int
        The number of atoms to select.
    stochastic: bool
        Whether to perform stochastic CUR selection.
    random_seed: int
        The seed for the random number generator.

    Returns
    -------
    list of ase.Atoms
        The selected atoms.

    Notes
    -----
    This function calculates the descriptor vector for each atom,
    then performs CUR selection on the resulting vectors.

    References
    ----------
    *    Title: Research data supporting "De novo exploration and self-guided learning of potential-energy surfaces"
    *    Script: select_by_descriptor.py
    *    Author: Noam Bernstein, Gábor Csányi and Volker L. Deringer
    *    Date 11/10/2019
    *    Availability: https://www.repository.cam.ac.uk/items/3aff252b-a583-4e7c-afc9-9dc1540cc37e
    *    License: Attribution 4.0 International (CC BY 4.0) license.
    """
    if random_seed is not None:
        np.random.seed(random_seed)

    if isinstance(atoms[0], list):
        print("flattening")
        fatoms = flatten(atoms, recursive=True)
    else:
        fatoms = atoms

    num_workers = min(len(fatoms), os.cpu_count() or 1)

    descriptor_worker = partial(
        parallel_calc_descriptor_vec, selected_descriptor=selected_descriptor
    )
    with threadpool_limits(limits=1), Pool(processes=num_workers) as pool:
        # TODO: implement argument for number of cores throughout
        ats = pool.map(descriptor_worker, fatoms)

    if isinstance(ats, list) & (len(ats) != 0):
        at_descs = np.array([at.info["descriptor_vec"] for at in ats]).T
        m = (
            np.matmul((np.squeeze(at_descs)).T, np.squeeze(at_descs)) ** kernel_exp
            if kernel_exp > 0.0
            else at_descs
        )

        def descriptor_svd(at_descs, num, do_vectors="vh"):
            def mv(v):
                return np.dot(at_descs, v)

            def rmv(v):
                return np.dot(at_descs.T, v)

            A = LinearOperator(at_descs.shape, matvec=mv, rmatvec=rmv, matmat=mv)
            return svds(A, k=num, return_singular_vectors=do_vectors)

        (_, _, vt) = descriptor_svd(
            m, min(max(1, int(select_nums / 2)), min(m.shape) - 1)
        )
        c_scores = np.sum(vt**2, axis=0) / vt.shape[0]
        if stochastic:
            selected = sorted(
                np.random.choice(
                    range(len(ats)), size=select_nums, replace=False, p=c_scores
                )
            )
        else:
            selected = sorted(np.argsort(c_scores)[-select_nums:])

        selected_atoms = [ats[i] for i in selected]

        for at in selected_atoms:
            del at.info["descriptor_vec"]

        return selected_atoms

    return None


def boltzhist_cur_one_shot(
    atoms: list[Atoms] | list[list[Atoms]],
    descriptor: str,
    isolated_atom_energies: dict,
    bolt_frac: float = 0.1,
    bolt_max_num: int = 3000,
    cur_num: int = 100,
    kernel_exp: float = 4,
    kt: float = 0.3,
    energy_label: str = "energy",
    pressures: list[float] | list[list[float]] | None = None,
    random_seed: int = None,
) -> list | None:
    """
    Sample atoms from a list according to boltzmann energy weighting and CUR diversity.

    Parameters
    ----------
    atoms: list[ase.Atoms] or list[list[ase.Atoms]]
        The atoms from which to select. If this is a list of lists, it is flattened.
    descriptor: str
        The quippy SOAP descriptor string for CUR.
    isolated_atom_energies: dict
        Dictionary of isolated energy values for species. Required for 'boltzhist_cur'
        selection method.
    bolt_frac: float
        The fraction to control the flat Boltzmann selection number.
    bolt_max_num: int
        The maximum number of atoms to select by Boltzmann-weighted flat histogram.
    cur_num: int
        The number of atoms to select by CUR.
    kernel_exp: float
        The exponent for the dot-product SOAP kernel.
    kt: float
        The product of the Boltzmann constant and the temperature, in eV.
    energy_label: str
        The label for the energy property in the atoms.
    pressures: list[float] or list[list[float]]
        The pressures at which the atoms have been optimized, in GPa.
    random_seed : int
        The seed for the random number generator.

    Returns
    -------
    list of ase.Atoms
        The selected atoms. These are copies of the atoms in the input list.

    Notes
    -----
    The algorithm uses a combination of CUR selection and Boltzmann weighting
    to select the atoms with diversity and low energy.

    References
    ----------
    *    Title: Research data supporting "De novo exploration and self-guided learning of potential-energy surfaces"
    *    Script: select_enthalpy_flat_histogram.py
    *    Author: Noam Bernstein, Gábor Csányi and Volker L. Deringer
    *    Date 11/10/2019
    *    Availability: https://www.repository.cam.ac.uk/items/3aff252b-a583-4e7c-afc9-9dc1540cc37e
    *    License: Attribution 4.0 International (CC BY 4.0) license.
    """
    if random_seed is not None:
        np.random.seed(random_seed)

    if isinstance(atoms[0], list):
        print("flattening")
        fatoms = flatten(atoms, recursive=True)
    else:
        fatoms = atoms

    if pressures is None:
        logging.info("Pressures not supplied, attempting to use pressure in atoms dict")

        try:
            ps = np.array([at.info["RSS_applied_pressure"] for at in fatoms])
        except RuntimeError:
            print("No pressures, so can't Boltzmann weight")

    else:
        ps = flatten_list(pressures)

    enthalpies = []

    at_ids = [atom.get_atomic_numbers() for atom in fatoms]

    if energy_label == "energy":
        formation_energies = []
        for ct, atom in enumerate(fatoms):
            if "energy" in atom.info:
                formation_energy = atom.info["energy"] - sum(
                    [isolated_atom_energies[j] for j in at_ids[ct]]
                )
            else:
                formation_energy = atom.get_potential_energy() - sum(
                    [isolated_atom_energies[j] for j in at_ids[ct]]
                )
            formation_energies.append(formation_energy)
        formation_energies = np.array(formation_energies)

    else:
        formation_energies = np.array(
            [
                atom.info[energy_label]
                - sum([isolated_atom_energies[j] for j in at_ids[ct]])
                for ct, atom in enumerate(fatoms)
            ]
        )

    for i, at in enumerate(fatoms):
        enthalpy = (formation_energies[i] + at.get_volume() * ps[i] * GPa) / len(at)
        enthalpies.append(enthalpy)

    enthalpies = np.array(enthalpies)
    min_H = np.min(enthalpies)
    config_prob = []
    histo = np.histogram(enthalpies)
    for H in enthalpies:
        bin_i = np.searchsorted(histo[1][1:], H, side="right")
        if bin_i == len(histo[1][1:]):
            bin_i = bin_i - 1
        p = 1.0 / histo[0][bin_i] if histo[0][bin_i] > 0.0 else 0.0
        if kt > 0.0:
            p *= np.exp(-(H - min_H) / kt)
        config_prob.append(p)

    select_num = round(bolt_frac * len(fatoms))

    select_num = select_num if select_num < bolt_max_num else bolt_max_num

    config_prob = np.array(config_prob)
    selected_bolt_ats = []
    for _ in range(select_num):
        config_prob /= np.sum(config_prob)
        cumul_prob = np.cumsum(config_prob)
        rv = np.random.uniform()
        config_i = np.searchsorted(cumul_prob, rv)
        selected_bolt_ats.append(fatoms[config_i])
        # remove from config_prob by converting to list
        config_prob = np.delete(config_prob, config_i)
        # remove from other lists
        del fatoms[config_i]
        enthalpies = np.delete(enthalpies, config_i)

    if cur_num < select_num:
        selected_atoms = cur_select(
            atoms=selected_bolt_ats,
            selected_descriptor=descriptor,
            kernel_exp=kernel_exp,
            select_nums=cur_num,
            stochastic=True,
            random_seed=random_seed,
        )
    else:
        selected_atoms = selected_bolt_ats

    return selected_atoms


def boltzhist_cur_dual_iter(
    atoms: list[Atoms] | list[list[Atoms]],
    descriptor: str,
    isolated_atom_energies: dict,
    bolt_frac: float = 0.1,
    bolt_max_num: int = 3000,
    cur_num: int = 100,
    kernel_exp: float = 4,
    kt: float = 0.3,
    energy_label: str = "energy",
    pressures: list[list[float]] | None = None,
    random_seed: int = None,
) -> list | None:
    """
    Execute sampling with two iterations.

    Each iteration includes a Boltzmann flat histogram in enthalpy
    followed by a CUR process.

    Parameters
    ----------
    atoms: list[ase.Atoms] or list[list[ase.Atoms]]
        The atoms from which to select. If this is a list of lists, it is flattened.
    descriptor: str
        The quippy SOAP descriptor string for CUR.
    isolated_atom_energies: dict
        Dictionary of isolated energy values for species. Required for 'boltzhist_cur'
        selection method. Default is None.
    bolt_frac: float
        The fraction to control the flat Boltzmann selection number.
    bolt_max_num: int
        The maximum number of atoms to select by Boltzmann-weighted flat histogram.
    cur_num: int
        The number of atoms to select by CUR.
    kernel_exp: float
        The exponent for the dot-product SOAP kernel.
    kt: float
        The product of the Boltzmann constant and the temperature, in eV.
    energy_label: str
        The label for the energy property in the atoms.
    pressures: list[float] or list[list[float]]
        The pressures at which the atoms have been optimized, in GPa.
    random_seed : int
        The seed for the random number generator.

    Returns
    -------
    list of ase.Atoms
        The selected atoms. These are copies of the atoms in the input list.

    Notes
    -----
    This function selects the most diverse atoms based on the chosen algorithm.
    The algorithm uses a combination of CUR selection and Boltzmann weighting to select the atoms.
    """
    atom_minima = [ats[-1] for ats in atoms]
    minima_indices = [ats[-1].info["unique_starting_index"] for ats in atoms]
    pressure_minima = None if pressures is None else [p[-1] for p in pressures]

    selected_minima = boltzhist_cur_one_shot(
        atoms=atom_minima,
        descriptor=descriptor,
        isolated_atom_energies=isolated_atom_energies,
        bolt_frac=bolt_frac,
        bolt_max_num=bolt_max_num,
        cur_num=cur_num,
        kernel_exp=kernel_exp,
        kt=kt,
        energy_label=energy_label,
        pressures=pressure_minima,
        random_seed=random_seed,
    )

    if selected_minima is None:
        raise ValueError(
            "The structures obtained from the first bcur sampling cannot be None."
        )

    selected_minima_indices = [
        at.info["unique_starting_index"] for at in selected_minima
    ]
    selected__indices = [minima_indices.index(item) for item in selected_minima_indices]
    selected_trajs = [atoms[i] for i in selected__indices]
    selected_trajs_pressure = (
        None if pressures is None else [pressures[j] for j in selected__indices]
    )

    return boltzhist_cur_one_shot(
        atoms=selected_trajs,
        descriptor=descriptor,
        isolated_atom_energies=isolated_atom_energies,
        bolt_frac=bolt_frac,
        bolt_max_num=bolt_max_num,
        cur_num=cur_num,
        kernel_exp=kernel_exp,
        kt=kt,
        energy_label=energy_label,
        pressures=selected_trajs_pressure,
        random_seed=random_seed,
    )


def convexhull_cur(
    atoms: list[Atoms],
    descriptor: str,
    bolt_frac: float = 0.1,
    bolt_max_num: int = 3000,
    cur_num: int = 100,
    kernel_exp: float = 4,
    kt: float = 0.5,
    energy_label: str = "REF_energy",
    isolated_atom_energies: dict = None,
    element_order: list | None = None,
    scheme: str = "linear-hull",
) -> list | None:
    """
    Sample atoms from a list according to Boltzmann energy weighting relative to convex hull and CUR diversity.

    Parameters
    ----------
    atoms: list of ase.Atoms
        The atoms for which to perform CUR selection.
    bolt_frac: float
        The fraction to control the proportion of atoms kept
        during the Boltzmann selection step.
    bolt_max_num: int
        The maximum number of atoms to select by Boltzmann flat
        histogram.
    cur_num: int
        The number of atoms to select by CUR.
    kernel_exp: float
        The kernel exponent to use in the calculation.
    kt: float
        The product of the Boltzmann constant and the temperature,
        in eV.
    energy_label: str
        The label for the energy property in the atoms.
    descriptor: str, optional
        The quip descriptor string to use for the calculation.
    isolated_atom_energies: dict, optional
        The isolated atom energies for each element in the system.
    element_order: list of str, optional
        The order of elements for the isolated atom energies.
    scheme: str, optional
        The scheme to use for the convex hull calculation.
        Default is 'linear-hull' (2D E,V hull).
        For 2-component systems with varying stoichiometry,
        use 'volume-stoichiometry' (>=3D E,V,mole-fraction hull).
        TODO: need to generalise this to ND hulls for mcp systems.
        GST good test case.

    Returns
    -------
    list of ase.Atoms
        The selected atoms.

    Notes
    -----
    This function calculates the descriptor vector for each atom,
    then performs CUR selection on the resulting vectors.
    The selection is based on the convex hull of the vectors.
    """
    if isinstance(atoms[0], list):
        print("flattening")
        fatoms = flatten(atoms, recursive=True)
    else:
        fatoms = atoms

    if isolated_atom_energies is None:
        raise KeyError("isolated_atom_energies must be supplied for convexhull_cur")

    if scheme == "linear-hull":
        hull, p = get_convex_hull(fatoms, energy_name=energy_label)
        des = np.array(
            [
                get_e_distance_to_hull(hull, at, energy_name=energy_label)
                for at in fatoms
            ]
        )

    elif scheme == "volume-stoichiometry":
        points = label_stoichiometry_volume(
            fatoms,
            isolated_atom_energies=isolated_atom_energies,
            energy_name=energy_label,
            element_order=element_order,
        )
        hull = calculate_hull_nd(points)

        des = np.array(
            [
                get_e_distance_to_hull_nd(
                    hull,
                    at,
                    isolated_atom_energies=isolated_atom_energies,
                    energy_name=energy_label,
                    element_order=element_order,
                )
                for at in fatoms
            ]
        )
        print("it will be coming soon!")

    else:
        raise ValueError(
            'scheme must be either "linear-hull" or "volume-stoichiometry"'
        )

    histo = np.histogram(des)
    config_prob = []
    min_ec = np.min(des)

    for ec in des:
        bin_i = np.searchsorted(histo[1][1:], ec, side="right")
        if bin_i == len(histo[1][1:]):
            bin_i = bin_i - 1
        p = 1.0 / histo[0][bin_i] if histo[0][bin_i] > 0.0 else 0.0
        if kt > 0.0:
            p *= np.exp(-(ec - min_ec) / kt)
        config_prob.append(p)

    select_num = round(bolt_frac * len(fatoms))

    select_num = select_num if select_num < bolt_max_num else bolt_max_num

    config_prob = np.array(config_prob)
    selected_bolt_ats = []
    for _ in range(select_num):
        config_prob /= np.sum(config_prob)
        cumul_prob = np.cumsum(config_prob)  # cumulate prob
        rv = np.random.uniform()
        config_i = np.searchsorted(cumul_prob, rv)
        selected_bolt_ats.append(fatoms[config_i])
        # remove from config_prob by converting to list
        config_prob = np.delete(config_prob, config_i)
        # remove from other lists
        del fatoms[config_i]
        des = np.delete(des, config_i)

    # implement CUR
    if cur_num < select_num:
        selected_atoms = cur_select(
            atoms=selected_bolt_ats,
            selected_descriptor=descriptor,
            kernel_exp=kernel_exp,
            select_nums=cur_num,
            stochastic=True,
        )
    else:
        selected_atoms = selected_bolt_ats

    return selected_atoms


def data_distillation(
    dft_ref_dir: str, force_max: float, force_label: str
) -> list[Atom | Atoms]:
    """
    For data distillation.

    Parameters
    ----------
    dft_ref_dir: str
        DFT reference data directory.
    force_max: float
        Maximally allowed force.
    force_label: str
        The label for the force property in the atoms.

    Returns
    -------
    atoms_distilled:
        List of distilled atoms.

    """
    atoms = ase.io.read(dft_ref_dir, index=":")

    atoms_distilled = []
    for at in atoms:
        forces = np.abs(at.arrays[force_label])
        f_component_max = np.max(forces)

        if f_component_max < force_max:
            atoms_distilled.append(at)

    logging.warning(
        f"After distillation, there are still {len(atoms_distilled)} data points remaining."
    )

    return atoms_distilled


def stratified_dataset_split(
    atoms: Atoms, split_ratio: float, energy_label: str
) -> tuple[
    list[Atom | Atoms]
    | list[Atom | Atoms | list[Atom | Atoms] | list[Atom | Atoms | list]],
    list[Atom | Atoms | list[Atom | Atoms] | list[Atom | Atoms | list]],
]:
    """
    Split the dataset.

    Parameters
    ----------
    atoms: Atoms
        ASE Atoms object
    split_ratio: float
        Parameter to divide the training set and the test set.
    energy_label: str
        The label for the energy property in the atoms.

    Returns
    -------
    train_structures, test_structures:
        Split-up datasets of train structures and test structures.

    """
    atom_bulk = []
    atom_isolated_and_dimer = []
    for at in atoms:
        if (
            at.info["config_type"] != "dimer"
            and at.info["config_type"] != "IsolatedAtom"
        ):
            atom_bulk.append(at)
        else:
            atom_isolated_and_dimer.append(at)

    if len(atoms) != len(atom_bulk):
        atoms = atom_bulk

    # Need this try except block because the energy label is not present as info
    try:
        average_energies = np.array(
            [atom.info[energy_label] / len(atom) for atom in atoms]
        )
    except KeyError:
        average_energies = np.array(
            [atom.get_potential_energy() / len(atom) for atom in atoms]
        )
    # sort by energy
    sorted_indices = np.argsort(average_energies)
    atoms = [atoms[i] for i in sorted_indices]
    average_energies = average_energies[sorted_indices]

    stratified_average_energies = pd.qcut(average_energies, q=2, labels=False)
    split = StratifiedShuffleSplit(n_splits=1, test_size=split_ratio, random_state=42)

    for train_index, test_index in split.split(atoms, stratified_average_energies):
        train_structures = [atoms[i] for i in train_index]
        test_structures = [atoms[i] for i in test_index]

    if atom_isolated_and_dimer:
        train_structures = atom_isolated_and_dimer + train_structures

    return train_structures, test_structures


def create_soap_descriptor(
    soap_paras: dict[str, int | float | str], n_species: int, species_Z: str
) -> str:
    """
    Generate a SOAP descriptor string based on the given parameters.

    Parameters
    ----------
    soap_paras:
        A dictionary containing SOAP parameters.
    n_species:
        The number of species.
    species_Z:
        A string representing species Z.
    """
    return (
        "soap l_max="
        + str(soap_paras["l_max"])
        + " n_max="
        + str(soap_paras["n_max"])
        + " atom_sigma="
        + str(soap_paras["atom_sigma"])
        + " cutoff="
        + str(soap_paras["cutoff"])
        + " n_species="
        + str(n_species)
        + " species_Z="
        + species_Z
        + " cutoff_transition_width="
        + str(soap_paras["cutoff_transition_width"])
        + " average="
        + str(soap_paras["average"])
    )


def flatten_list(input_list: list | list[list]) -> list:
    """Flatten a nested list into a single list if necessary."""
    if (
        isinstance(input_list, list)
        and len(input_list) > 0
        and isinstance(input_list[0], list)
    ):
        return list(chain.from_iterable(input_list))

    return input_list


def handle_rss_trajectory(
    traj_path, remove_traj_files
) -> tuple[list[list], list[list]]:
    """
    Handle trajectory and associated information.

    Parameters
    ----------
    traj_path: list | None
        List of dictionaries containing trajectory information.
        Each dictionary should have keys 'traj_path' and 'pressure'.
        If None, an empty list will be used.
    remove_traj_files: bool
        Whether to remove the directories containing trajectory files
        after processing them. Default is False.

    Returns
    -------
    tuple:
        atoms: list
            List of ASE Atoms objects read from the trajectory files.
        pressures: list
            List of pressure values corresponding to the atoms.
    """
    atoms = []
    pressures = []
    traj_path = [] if traj_path is None else flatten_list(traj_path)
    traj_dirs = []

    if all(i is None for i in traj_path):
        raise ValueError("No valid trajectory path was obtained!")

    for traj in traj_path:
        if traj is not None and Path(traj).exists():
            print("Processing trajectory:", traj)
            at = ase.io.read(traj, index=":")
            atoms.append(at)
            pressure = [i.info["RSS_applied_pressure"] for i in at]
            pressures.append(pressure)
            traj_dirs.append(os.path.dirname(traj))

    if remove_traj_files and traj_dirs:
        traj_dirs = list(set(traj_dirs))
        for dir_path in traj_dirs:
            if os.path.exists(dir_path) and os.path.isdir(dir_path):
                for root, _, files in os.walk(dir_path):
                    for name in files:
                        if "RSS_relax_results" in name:
                            file_path = os.path.join(root, name)
                            os.remove(file_path)

    return atoms, pressures



"""RSS (random structure searching) flow for exploring and learning potential energy surfaces from scratch."""

from dataclasses import dataclass, field

from atomate2.forcefields.jobs import ForceFieldStaticMaker
from atomate2.vasp.jobs.base import BaseVaspMaker
from atomate2.vasp.jobs.core import StaticMaker
from atomate2.vasp.sets.core import StaticSetGenerator
from jobflow import Flow, Maker, Response, job

from autoplex.auto.rss.jobs import do_rss_iterations, initial_rss
from autoplex.misc.castep.jobs import CastepStaticMaker
from autoplex.settings import RssConfig


@dataclass
class RssMaker(Maker):
    """
    Maker to set up and run RSS for exploring and learning potential-energy surfaces (from scratch).

    Parameters
    ----------
    name: str
        Name of the flow.
    rss_config: RssConfig
        Pydantic model that defines the setup parameters for the whole RSS workflow.
        If not explicitly set, the defaults from 'autoplex.settings.RssConfig' will be used.
    static_energy_maker: BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker
        Maker for static energy jobs: either BaseVaspMaker (VASP-based) or CastepStaticMaker (CASTEP-based) or
        ForceFieldStaticMaker (force field-based). Defaults to StaticMaker (VASP-based).
    static_energy_maker_isolated_atoms: BaseVaspMaker | ForceFieldStaticMaker | None
        Maker for static energy jobs of isolated atoms: either BaseVaspMaker (VASP-based) or CastepStaticMaker
        (CASTEP-based) or ForceFieldStaticMaker (force field-based) or None. If set to `None`, the parameters
        from `static_energy_maker` will be used as the default for isolated atoms. In this case,
        if `static_energy_maker` is a `StaticMaker`, all major settings will be inherited,
        except that `kspacing` will be automatically set to 100 to enforce a Gamma-point-only calculation.
        This is typically suitable for single-atom systems. Default is None. If a non-`StaticMaker` maker
        is used here, its output must include a `dir_name` field to ensure compatibility with downstream workflows.

    """

    name: str = "ml-driven rss"
    rss_config: RssConfig = field(default_factory=lambda: RssConfig())
    static_energy_maker: BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker = (
        field(
            default_factory=lambda: StaticMaker(
                input_set_generator=StaticSetGenerator(
                    user_incar_settings={
                        "ADDGRID": "True",
                        "ENCUT": 520,
                        "EDIFF": 1e-06,
                        "ISMEAR": 0,
                        "SIGMA": 0.01,
                        "PREC": "Accurate",
                        "ISYM": None,
                        "KSPACING": 0.2,
                        "NPAR": 8,
                        "LWAVE": "False",
                        "LCHARG": "False",
                        "ENAUG": None,
                        "GGA": None,
                        "ISPIN": None,
                        "LAECHG": None,
                        "LELF": None,
                        "LORBIT": None,
                        "LVTOT": None,
                        "NSW": None,
                        "SYMPREC": None,
                        "NELM": 100,
                        "LMAXMIX": None,
                        "LASPH": None,
                        "AMIN": None,
                    }
                ),
                run_vasp_kwargs={"handlers": ()},
            )
        )
    )
    static_energy_maker_isolated_atoms: (
        BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker | None
    ) = None

    @job
    def make(self, **kwargs):
        """
        Make a rss workflow using the specified configuration file and additional keyword arguments.

        Parameters
        ----------
        kwargs: dict, optional
            Additional optional keyword arguments to customize the job execution.

        Keyword Arguments
        -----------------
        tag: str
            Tag of systems. It can also be used for setting up elements and stoichiometry.
            For example, the tag of 'SiO2' will be recognized as a 1:2 ratio of Si to O and
            passed into the parameters of buildcell. However, note that this will be overwritten
            if the stoichiometric ratio of elements is defined in the 'cell_seed_paths' or 'buildcell_options'.
        train_from_scratch: bool
            If True, it starts the workflow from scratch.
            If False, it resumes from a previous state.
        resume_from_previous_state: dict | None
            A dictionary containing the state information required to resume a previously interrupted
            or saved RSS workflow. When 'train_from_scratch' is set to False, this parameter is mandatory
            for the workflow to pick up from a saved state.Expected keys within this dictionary are as follows

            - 'test_error': float, The test error from the last completed training step.
            - 'pre_database_dir': str, Path to the directory containing the pre-existing database for resuming.
            - 'mlip_path': str, Path to the file of a previous MLIP model.
            - 'isolated_atom_energies': dict, A dictionary with isolated atom energy values mapped to atomic numbers.

        generated_struct_numbers: list[int]
            Expected number of generated randomized unit cells by buildcell.
        cell_seed_paths: list[str]
            A list of paths to the custom buildcell control files, which ends with '.cell'. If these files exist,
            the buildcell_options argument will no longer take effect.
        buildcell_options: list[dict] | None
            Customized parameters for buildcell. Default is None.
        fragment: Atoms | list[Atoms] | None
            Fragment(s) for random structures, e.g., molecules, to be placed individually intact.
            atoms.arrays should have a 'fragment_id' key with unique identifiers for each fragment if in same Atoms.
            atoms.cell must be defined (e.g., Atoms.cell = np.eye(3)*20).
        fragment_numbers: list[str] | None
            Numbers of each fragment to be included in the random structures. Defaults to 1 for all specified.
        num_processes_buildcell: int
            Number of processes to use for parallel computation during buildcell generation.
        num_of_initial_selected_structs: list[int] | None
            Number of structures to be sampled directly from the buildcell-generated randomized cells.
        num_of_rss_selected_structs: int
            Number of structures to be selected from each RSS iteration.
        initial_selection_enabled: bool
            If true, sample structures from initially generated randomized cells using CUR.
        rss_selection_method: str
            Method for selecting samples from the RSS trajectories:
            Boltzmann flat histogram in enthalpy first, then CUR.
            Options are as follows

            - 'bcur1s': Execute bcur with one shot (1s)
            - 'bcur2i': Execute bcur with two iterations (2i)

        bcur_params: dict | None
            Parameters for Boltzmann CUR selection. The default dictionary includes following keys

            soap_paras: dict
                SOAP descriptor parameters dict with following acceptable keys

            - 'l_max': int, Maximum degree of spherical harmonics (default 12).
            - 'n_max': int, Maximum number of radial basis functions (default 12).
            - 'atom_sigma': float, Width of Gaussian smearing (default 0.0875).
            - 'cutoff': float, Radial cutoff distance (default 10.5).
            - 'cutoff_transition_width': float, Width of the transition region (default 1.0).
            - 'zeta': float,Exponent for dot-product SOAP kernel (default 4.0).
            - 'average': bool, Whether to average the SOAP vectors (default True).
            - 'species': bool, Whether to consider species information (default True).

            kb_temp: float
                Temperature in eV for Boltzmann weighting (default 0.3).
            frac_of_bcur: float
                Fraction of Boltzmann CUR selections (default 0.8).
            bolt_max_num: int
                Maximum number of Boltzmann selections (default 3000).
            kernel_exp: float
                Exponent for the kernel (default 4.0).
            energy_label: str
                Label for the energy data (default 'energy').
        random_seed: int | None
            A seed to ensure reproducibility of CUR selection. Default is None.
        include_isolated_atom: bool
            If true, perform single-point calculations for isolated atoms.
        isolatedatom_box: list[float]
            List of the lattice constants for an isolated atom configuration.
        e0_spin: bool
            If true, include spin polarization in isolated atom and dimer calculations. Default is False.
        include_dimer: bool
            If true, perform single-point calculations for dimers only once. Default is False.
        dimer_box: list[float]
            The lattice constants of a dimer box.
        dimer_range: list[float]
            Range of distances for dimer calculations.
        dimer_num: int
            Number of different distances to consider for dimer calculations. Default is 21.
        custom_incar: dict | None
            Dictionary of custom DFT input parameters. If provided, will update the
            default parameters. Default is None.
        custom_potcar: dict | None
            Dictionary of POTCAR settings to update. Keys are element symbols, values are the desired POTCAR labels.
            Default is None.
        dft_ref_file: str
            Reference file for DFT data. Default is 'dft_ref.extxyz'.
        config_types: list[str]
            Configuration types for the DFT calculations. Default is None.
        rss_group: list[str] | str
            Group name for RSS to setting up regularization.
        test_ratio: float
            The proportion of the test set after splitting the data. The value is allowed to be set to 0;
            in this case, the testing error would not be meaningful anymore.
        regularization: bool
            If True, apply regularization. This only works for GAP to date. Default is False.
        retain_existing_sigma: bool
            Whether to keep the current sigma values for specific configuration types.
            If set to True, existing sigma values for specific configurations will remain unchanged.
        scheme: str
            Method to use for regularization. Options are

            - 'linear_hull': for single-composition system, use 2D convex hull (E, V)
            - 'volume-stoichiometry': for multi-composition system, use 3D convex hull of (E, V, mole-fraction)

        reg_minmax: list[tuple]
            List of tuples of (min, max) values for energy, force, virial sigmas for regularization.
        distillation: bool
            If true, apply data distillation. Default is True.
        force_max: float | None
            Maximum force value to exclude structures. Default is 50.
        force_label: str | None
            The label of force values to use for distillation. Default is 'REF_forces'.
        pre_database_dir: str | None
            Directory where the previous database was saved.
        mlip_type: str
            Choose one specific MLIP type to be fitted: 'GAP' | 'J-ACE' | 'NEQUIP' | 'M3GNET' | 'MACE'.
            Default is 'GAP'.
        ref_energy_name: str
            Reference energy name. Default is 'REF_energy'.
        ref_force_name: str
            Reference force name. Default is 'REF_forces'.
        ref_virial_name: str
            Reference virial name. Default is 'REF_virial'.
        auto_delta: bool
            If true, apply automatic determination of delta for GAP terms. Default is False.
        num_processes_fit: int
            Number of processes used for fitting. Default is 1.
        device_for_fitting: str
            Device to be used for model fitting, either "cpu" or "cuda".
        scalar_pressure_method: str
            Method for adding external pressures. Acceptable options are as follows

            - 'exp': Applies pressure using an exponential distribution.
            - 'uniform': Applies pressure using a uniform distribution.

        scalar_exp_pressure: float
            Scalar exponential pressure. Default is 100.
        scalar_pressure_exponential_width: float
            Width for scalar pressure exponential. Default is 0.2.
        scalar_pressure_low: float
            Low limit for scalar pressure. Default is 0.
        scalar_pressure_high: float
            High limit for scalar pressure. Default is 50.
        max_steps: int
            Maximum number of steps for relaxation. Default is 200.
        force_tol: float
            Force residual tolerance for relaxation. Default is 0.05.
        stress_tol: float
            Stress residual tolerance for relaxation. Default is 0.05.
        hookean_repul: bool
            If true, apply Hookean repulsion. Default is False.
        hookean_paras: dict[tuple[int, int], tuple[float, float]] | None
            Parameters for Hookean repulsion as a dictionary of tuples. Default is None.
        keep_symmetry: bool
            If true, preserve symmetry during relaxation. Default is False.
        remove_traj_files: bool
            If true, remove all trajectory files raised by RSS to save memory
        num_processes_rss: int
            Number of processes used for running RSS. Default is 1.
        device_for_rss: str
            Specify device to use "cuda" or "cpu" for running RSS. Default is "cpu".
        stop_criterion: float
            Convergence criterion for stopping RSS iterations. Default is 0.01.
        max_iteration_number: int
            Maximum number of RSS iterations to perform. Default is 25.
        num_groups: int
            Number of structure groups, used for assigning tasks across multiple nodes.
            For example, if there are 10,000 trajectories to relax and 'num_groups=10',
            the trajectories will be divided into 10 groups and 10 independent jobs will be created,
            with each job handling 1,000 trajectories.
        initial_kb_temp: float
            Initial temperature (in eV) for Boltzmann sampling. Default is 0.3.
        current_iter_index: int
            Index for the current RSS iteration. Default is 1.
        **fit_kwargs:
            Additional keyword arguments for the MLIP fitting process.

        Returns
        -------
        dict:
            A dictionary with following information

            - 'test_error': float, The test error of the fitted MLIP.
            - 'pre_database_dir': str, The directory of the latest RSS database.
            - 'mlip_path': str | Path, Path to the latest fitted MLIP.
            - 'isolated_atom_energies': dict, The isolated energy values.
            - 'current_iter': int, The current iteration index.
            - 'kb_temp': float, The temperature (in eV) for Boltzmann sampling.
        """
        default_config = self.rss_config.model_copy(deep=True)
        if kwargs:
            default_config.update_parameters(kwargs)

        config_params = default_config.model_dump(by_alias=True, exclude_none=True)

        # Extract MLIP hyperparameters from the config_params
        mlip_hypers = config_params["mlip_hypers"][config_params["mlip_type"]]
        del config_params["mlip_hypers"]
        config_params.update(mlip_hypers)

        self._process_hookean_paras(config_params)

        if "train_from_scratch" not in config_params:
            raise ValueError(
                "'train_from_scratch' must be set in the configuration file or passed as a keyword argument!!"
            )

        rss_flow = []

        if config_params["train_from_scratch"]:
            initial_exclude_keys = [
                "train_from_scratch",
                "resume_from_previous_state",
                "config_types",
                "rss_group",
                "num_of_rss_selected_structs",
                "rss_selection_method",
                "scalar_pressure_method",
                "scalar_exp_pressure",
                "scalar_pressure_exponential_width",
                "scalar_pressure_low",
                "scalar_pressure_high",
                "max_steps",
                "force_tol",
                "stress_tol",
                "stop_criterion",
                "max_iteration_number",
                "num_groups",
                "initial_kb_temp",
                "current_iter_index",
                "hookean_repul",
                "hookean_paras",
                "keep_symmetry",
                "remove_traj_files",
                "num_processes_rss",
                "device_for_rss",
            ]
            initial_params = {
                k: v for k, v in config_params.items() if k not in initial_exclude_keys
            }
            initial_params.update(
                {
                    "config_type": config_params["config_types"][0],
                    "rss_group": config_params["rss_group"][0],
                }
            )
            initial_rss_job = initial_rss(
                static_energy_maker=self.static_energy_maker,
                static_energy_maker_isolated_atoms=self.static_energy_maker_isolated_atoms,
                **initial_params,
            )
            rss_flow.append(initial_rss_job)

        rss_group = config_params["rss_group"]
        config_types = config_params["config_types"]
        do_rss_group = rss_group[0] if len(rss_group) == 1 else rss_group[-1]
        rss_config_type = (
            config_types[0] if len(config_types) == 1 else config_types[1:]
        )

        rss_exclude_keys = [
            "train_from_scratch",
            "resume_from_previous_state",
            "pre_database_dir",
        ]

        rss_params = {
            k: v for k, v in config_params.items() if k not in rss_exclude_keys
        }
        rss_params.update(
            {
                "num_of_initial_selected_structs": None,
                "initial_selection_enabled": False,
                "rss_group": do_rss_group,
                "config_types": rss_config_type,
            }
        )

        if config_params["train_from_scratch"]:
            rss_params.update({"include_isolated_atom": False})
            rss_params.update({"include_dimer": False})

            do_rss_job = do_rss_iterations(
                input=initial_rss_job.output,
                static_energy_maker=self.static_energy_maker,
                static_energy_maker_isolated_atoms=self.static_energy_maker_isolated_atoms,
                **rss_params,
            )
        else:
            if "resume_from_previous_state" not in config_params:
                raise ValueError(
                    "The parameter 'resume_from_previous_state' must be specified when 'train_from_scratch' is False."
                )
            resume_from_previous_state = config_params["resume_from_previous_state"]
            do_rss_job = do_rss_iterations(
                input=resume_from_previous_state,
                static_energy_maker=self.static_energy_maker,
                static_energy_maker_isolated_atoms=self.static_energy_maker_isolated_atoms,
                **rss_params,
            )

        rss_flow.append(do_rss_job)

        return Response(replace=Flow(rss_flow), output=do_rss_job.output)

    @staticmethod
    def _process_hookean_paras(config):
        if "hookean_paras" in config and config["hookean_paras"] is not None:
            config["hookean_paras"] = {
                tuple(map(int, k.strip("()").split(", "))): tuple(v)
                for k, v in config["hookean_paras"].items()
            }
